import pytest
from dp_wizard_templates.code_template import Template


def test_def_too_long():
    def template(
        BEGIN,
        END,
    ):
        print(BEGIN, END)

    with pytest.raises(Exception, match=r"def and parameters should fit on one line"):
        Template(template)


def test_def_template():
    def template(BEGIN, END):
        print(BEGIN, END)

    assert (
        Template(template).fill_values(BEGIN="abc", END="xyz").finish()
        == "print('abc', 'xyz')"
    )


def test_fill_expressions():
    template = Template("No one VERB the ADJ NOUN!")
    filled = template.fill_expressions(
        VERB="expects",
        ADJ="Spanish",
        NOUN="Inquisition",
    ).finish()
    assert filled == "No one expects the Spanish Inquisition!"


def test_fill_expressions_missing_slot_in_template():
    template = Template("No one ... the ADJ NOUN!")
    with pytest.raises(Exception, match=r"No 'VERB' slot to fill with 'expects'"):
        template.fill_expressions(
            VERB="expects",
            ADJ="Spanish",
            NOUN="Inquisition",
        ).finish()


def test_fill_expressions_extra_slot_in_template():
    template = Template("No one VERB ARTICLE ADJ NOUN!")
    with pytest.raises(Exception, match=r"'ARTICLE' slot not filled"):
        template.fill_expressions(
            VERB="expects",
            ADJ="Spanish",
            NOUN="Inquisition",
        ).finish()


def test_fill_values():
    template = Template("assert [STRING] * NUM == LIST")
    filled = template.fill_values(
        STRING="🙂",
        NUM=3,
        LIST=["🙂", "🙂", "🙂"],
    ).finish()
    assert filled == "assert ['🙂'] * 3 == ['🙂', '🙂', '🙂']"


def test_fill_values_missing_slot_in_template():
    template = Template("assert [STRING] * ... == LIST")
    with pytest.raises(Exception, match=r"No 'NUM' slot to fill with '3'"):
        template.fill_values(
            STRING="🙂",
            NUM=3,
            LIST=["🙂", "🙂", "🙂"],
        ).finish()


def test_fill_values_extra_slot_in_template():
    template = Template("CMD [STRING] * NUM == LIST")
    with pytest.raises(Exception, match=r"'CMD' slot not filled"):
        template.fill_values(
            STRING="🙂",
            NUM=3,
            LIST=["🙂", "🙂", "🙂"],
        ).finish()


def test_fill_blocks():
    # "OK" is less than three characters, so it is not a slot.
    template = Template(
        """# MixedCase is OK

FIRST

with fake:
    SECOND
    if True:
        THIRD
""",
    )
    template.fill_blocks(
        FIRST="\n".join(f"import {i}" for i in "abc"),
        SECOND="\n".join(f"f({i})" for i in "123"),
        THIRD="\n".join(f"{i}()" for i in "xyz"),
    )
    assert (
        template.finish()
        == """# MixedCase is OK

import a
import b
import c

with fake:
    f(1)
    f(2)
    f(3)
    if True:
        x()
        y()
        z()
"""
    )


def test_fill_blocks_missing_slot_in_template_alone():
    template = Template("No block slot")
    with pytest.raises(Exception, match=r"No 'SLOT' slot"):
        template.fill_blocks(SLOT="placeholder").finish()


def test_fill_blocks_missing_slot_in_template_not_alone():
    template = Template("No block SLOT")
    with pytest.raises(
        Exception, match=r"Block slots must be alone on line; No 'SLOT' slot"
    ):
        template.fill_blocks(SLOT="placeholder").finish()


def test_fill_blocks_extra_slot_in_template():
    template = Template("EXTRA\nSLOT")
    with pytest.raises(Exception, match=r"'EXTRA' slot not filled"):
        template.fill_blocks(SLOT="placeholder").finish()


def test_fill_blocks_not_string():
    template = Template("SOMETHING")
    with pytest.raises(
        Exception,
        match=r"For SOMETHING in string template, expected string, not 123",
    ):
        template.fill_blocks(SOMETHING=123).finish()



#!/usr/bin/env python3
import subprocess
import re
from pathlib import Path


root_path = Path(__file__).parent.parent


def get_prev_version():
    """
    >>> len(get_prev_version().splitlines())
    1
    """
    return (root_path / "dp_wizard_templates" / "VERSION").read_text().strip()


def log_until(match):  # pragma: no cover
    lines = subprocess.check_output(["git", "log", "--oneline"], text=True).splitlines()
    if match is None:
        return lines
    head = []
    for line in lines:
        if match in line:
            break
        head.append(line)
    return head


def parse_log(lines):
    """
    >>> print(parse_log([
    ...     'abcd0000 bar! (#2)',
    ...     'abcd0001 foo? (#1)',
    ... ]))
    - bar! [#2](https://github.com/opendp/dp-wizard-templates/pull/2)
    - foo? [#1](https://github.com/opendp/dp-wizard-templates/pull/1)
    """
    output_lines = []
    for line in lines:
        line = re.sub(r"^\w+\s+", "", line)  # Remove hash
        line = re.sub(r"^\([^)]+\)\s+", "", line)  # Remove tag
        line = re.sub(
            r"\(#(\d+)\)",
            r"[#\1](https://github.com/opendp/dp-wizard-templates/pull/\1)",
            line,
        )
        output_lines.append(f"- {line}")
    return "\n".join(output_lines)


def main():  # pragma: no cover
    old_changelog_lines = (root_path / "CHANGELOG.md").read_text().splitlines()
    new_changelog_lines = []

    prev_version = get_prev_version()
    log_lines = log_until(prev_version)
    changelog_update = parse_log(log_lines)

    for line in old_changelog_lines:
        if prev_version in line:
            new_changelog_lines.append(changelog_update)
            new_changelog_lines.append("")
        new_changelog_lines.append(line)

    (root_path / "CHANGELOG.md").write_text("\n".join(new_changelog_lines))


if __name__ == "__main__":  # pragma: no cover
    main()



import inspect
import re
import black


def _get_body(func):

    source_lines = inspect.getsource(func).splitlines()
    first_line = source_lines[0]
    if not re.match(r"def \w+\((\w+(, \w+)*)?\):", first_line.strip()):
        # Parsing to AST and unparsing is a more robust option,
        # but more complicated.
        raise Exception(f"def and parameters should fit on one line: {first_line}")

    # The "def" should not be in the output,
    # and cleandoc handles the first line differently.
    source_lines[0] = ""
    body = inspect.cleandoc("\n".join(source_lines))
    body = re.sub(
        r"\s*#\s+type:\s+ignore\s*",
        "\n",
        body,
    )
    body = re.sub(
        r"\s*#\s+noqa:.+",
        "",
        body,
    )
    return body


class Template:
    def __init__(self, template, root=None):
        if root is not None:
            template_name = f"_{template}.py"
            template_path = root / template_name
            self._source = f"'{template_name}'"
            self._template = template_path.read_text()
        else:
            if callable(template):
                self._source = "function template"
                self._template = _get_body(template)
            else:
                self._source = "string template"
                self._template = template
        # We want a list of the initial slots, because substitutions
        # can produce sequences of upper case letters that could be mistaken for slots.
        self._initial_slots = self._find_slots()

    def _find_slots(self):
        # Slots:
        # - are all caps or underscores
        # - have word boundary on either side
        # - are at least three characters
        slot_re = r"\b[A-Z][A-Z_]{2,}\b"
        return set(re.findall(slot_re, self._template))

    def fill_expressions(self, **kwargs):
        """
        Fill in variable names, or dicts or lists represented as strings.
        """
        for k, v in kwargs.items():
            k_re = re.escape(k)
            self._template, count = re.subn(rf"\b{k_re}\b", str(v), self._template)
            if count == 0:
                raise Exception(
                    f"No '{k}' slot to fill with '{v}' in "
                    f"{self._source}:\n\n{self._template}"
                )
        return self

    def fill_values(self, **kwargs):
        """
        Fill in string or numeric values. `repr` is called before filling.
        """
        for k, v in kwargs.items():
            k_re = re.escape(k)
            self._template, count = re.subn(rf"\b{k_re}\b", repr(v), self._template)
            if count == 0:
                raise Exception(
                    f"No '{k}' slot to fill with '{v}' in "
                    f"{self._source}:\n\n{self._template}"
                )
        return self

    def fill_blocks(self, **kwargs):
        """
        Fill in code blocks. Slot must be alone on line.
        """
        for k, v in kwargs.items():
            if not isinstance(v, str):
                raise Exception(f"For {k} in {self._source}, expected string, not {v}")

            def match_indent(match):
                # This does what we want, but binding is confusing.
                return "\n".join(
                    match.group(1) + line for line in v.split("\n")  # noqa: B023
                )

            k_re = re.escape(k)
            self._template, count = re.subn(
                rf"^([ \t]*){k_re}$",
                match_indent,
                self._template,
                flags=re.MULTILINE,
            )
            if count == 0:
                base_message = (
                    f"No '{k}' slot to fill with '{v}' in "
                    f"{self._source}:\n\n{self._template}"
                )
                if k in self._template:
                    raise Exception(
                        f"Block slots must be alone on line; {base_message}"
                    )
                raise Exception(base_message)
        return self

    def finish(self, reformat=False) -> str:
        unfilled_slots = self._initial_slots & self._find_slots()
        if unfilled_slots:
            slots_str = ", ".join(sorted(f"'{slot}'" for slot in unfilled_slots))
            raise Exception(
                f"{slots_str} slot not filled "
                f"in {self._source}:\n\n{self._template}"
            )

        if reformat:
            self._template = black.format_str(self._template, mode=black.Mode())

        return self._template



import re
from pathlib import Path
import pytest
import json
from dp_wizard_templates.converters import (
    convert_py_to_nb,
    _clean_nb,
    convert_nb_to_html,
    ConversionException,
)


fixtures_path = Path(__file__).parent / "fixtures"


def norm_nb(nb_str):
    nb_str = re.sub(r'"id": "[^"]+"', '"id": "12345678"', nb_str)
    nb_str = re.sub(
        r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d+Z",
        "2024-01-01T00:00:00.000000Z",
        nb_str,
    )

    nb = json.loads(nb_str)
    nb["metadata"] = {k: v for k, v in nb["metadata"].items() if k == "title"}
    for cell in nb["cells"]:
        cell["metadata"] = {}

    return json.dumps(nb, indent=1)


def test_convert_py_to_nb():
    python_str = (fixtures_path / "fake.py").read_text()
    actual_nb_str = convert_py_to_nb(python_str, "Title!")
    expected_nb_str = (fixtures_path / "fake.ipynb").read_text()

    normed_actual_nb_str = norm_nb(actual_nb_str)
    normed_expected_nb_str = norm_nb(expected_nb_str)
    assert normed_actual_nb_str == normed_expected_nb_str


def test_convert_py_to_nb_execute():
    python_str = (fixtures_path / "fake.py").read_text()
    actual_nb_str = convert_py_to_nb(python_str, "Title!", execute=True)
    expected_nb_str = (fixtures_path / "fake-executed.ipynb").read_text()

    normed_actual_nb_str = norm_nb(actual_nb_str)
    normed_expected_nb_str = norm_nb(expected_nb_str)
    assert normed_actual_nb_str == normed_expected_nb_str


def test_clean_nb():
    # Trivial test just to get 100% branch coverage.
    nb = {"cells": []}
    assert nb == json.loads(_clean_nb(json.dumps(nb)))


def test_convert_py_to_nb_error():
    python_str = "Invalid python!"
    with pytest.raises(
        ConversionException,
        # There's more, but what's most important is that
        # the line with the error shows up in the message.
        match=(r"Invalid python!"),
    ):
        convert_py_to_nb(python_str, "Title!", execute=True, reformat=False)


def test_convert_nb_to_html():
    notebook = (fixtures_path / "fake-executed.ipynb").read_text()
    html = convert_nb_to_html(notebook)
    assert "[1]:" in html
    assert "<pre>4" in html



import subprocess
import re

import pytest

import dp_wizard_templates


tests = {
    "flake8 linting": "flake8 . --count --show-source --statistics",
    "pyright type checking": "pyright",
}


@pytest.mark.parametrize("cmd", tests.values(), ids=tests.keys())
def test_subprocess(cmd: str):
    result = subprocess.run(cmd, shell=True)
    assert result.returncode == 0, f'"{cmd}" failed'


def test_version():
    assert re.match(r"\d+\.\d+\.\d+", dp_wizard_templates.__version__)



from pathlib import Path
from tempfile import TemporaryDirectory
from dataclasses import dataclass
from sys import executable
import subprocess
import json
import nbformat
import nbconvert
import jupytext
import black


def _is_kernel_installed() -> bool:
    try:
        # This method isn't well documented, so it may be fragile.
        jupytext.kernels.kernelspec_from_language("python")  # type: ignore
        return True
    except ValueError:  # pragma: no cover
        return False


@dataclass(frozen=True)
class ConversionException(Exception):
    command: str
    stderr: str

    def __str__(self):
        return f"Script to notebook conversion failed: {self.command}\n{self.stderr})"


def convert_py_to_nb(
    python_str: str, title: str, execute: bool = False, reformat: bool = True
):
    """
    Given Python code as a string, returns a notebook as a string.
    Calls jupytext as a subprocess:
    Not ideal, but only the CLI is documented well.
    """
    with TemporaryDirectory() as temp_dir:
        if not _is_kernel_installed():
            subprocess.run(  # pragma: no cover
                [executable]
                + "-m ipykernel install --name kernel_name --user".split(" "),
                check=True,
            )

        temp_dir_path = Path(temp_dir)
        py_path = temp_dir_path / "input.py"
        if reformat:
            # Line length determined by PDF rendering.
            python_str = black.format_str(python_str, mode=black.Mode(line_length=74))
        py_path.write_text(python_str)

        argv = [executable] + "-m jupytext --from .py --to .ipynb --output -".split(" ")
        if execute:
            argv.append("--execute")
        argv.append(str(py_path.absolute()))  # type: ignore
        result = subprocess.run(argv, text=True, capture_output=True)
    if result.returncode != 0:
        # If there is an error, we want a copy of the file that will stay around,
        # outside the "with TemporaryDirectory()" block.
        # The command we show in the error message isn't exactly what was run,
        # but it should reproduce the error.
        debug_path = Path("/tmp/script.py")
        debug_path.write_text(python_str)
        argv.pop()
        argv.append(str(debug_path))  # type: ignore
        raise ConversionException(command=" ".join(argv), stderr=result.stderr)
    nb_dict = json.loads(result.stdout.strip())
    nb_dict["metadata"]["title"] = title
    return _clean_nb(json.dumps(nb_dict))


def _stable_hash(lines: list[str]):
    import hashlib

    return hashlib.sha1("\n".join(lines).encode()).hexdigest()[:8]


def _clean_nb(nb_json: str):
    """
    Given a notebook as a string of JSON, remove the coda and pip output.
    (The code may produce reports that we do need,
    but the code isn't actually interesting to end users.)
    """
    nb = json.loads(nb_json)
    new_cells = []
    for cell in nb["cells"]:
        if "pip install" in cell["source"][0]:
            cell["outputs"] = []
        # "Coda" may, or may not be followed by "\n".
        # Be flexible!
        if any(line.startswith("# Coda") for line in cell["source"]):
            break
        # Make ID stable:
        cell["id"] = _stable_hash(cell["source"])
        # Delete execution metadata:
        try:
            del cell["metadata"]["execution"]
        except KeyError:
            pass
        new_cells.append(cell)
    nb["cells"] = new_cells
    return json.dumps(nb, indent=1)


def convert_nb_to_html(python_nb: str, numbered=True):
    notebook = nbformat.reads(python_nb, as_version=4)
    exporter = nbconvert.HTMLExporter(
        template_name="lab",
        # The "classic" template's CSS forces large code cells on to
        # the next page rather than breaking, so use "lab" instead.
        #
        # If you want to tweak the CSS, enable this block and make changes
        # in nbconvert_templates/custom:
        #
        # template_name="custom",
        # extra_template_basedirs=[
        #     str((Path(__file__).parent / "nbconvert_templates").absolute())
        # ],
    )
    (body, _resources) = exporter.from_notebook_node(notebook)
    if not numbered:
        body = body.replace(
            "</head>",
            """
<style>
.jp-InputPrompt {display: none;}
</style>
</head>""",
        )
    return body



# # DP Wizard Templates
#
# [github](https://github.com/opendp/dp-wizard-templates) | [pypi](https://pypi.org/project/dp_wizard_templates/) | [docs](https://opendp.github.io/dp-wizard-templates) (this page)
#
# DP Wizard Templates helps you build Python code from templates
# which are themselves syntactically valid Python.
# Templates can be composed to generate entire notebooks.
#
# DP Wizard Templates relies on code inspection, so real working examples
# need to be in code, not in a notebook or a doctest.
# This documentation itself is rendered by the library.
#
# DP Wizard Templates was developed for
# [DP Wizard](https://github.com/opendp/dp-wizard),
# and that codebase remains a good place to look for further examples.
#
#
# ## Motivation
#
# Let's say you want to generate Python code programmatically,
# perhaps to demonstrate a workflow with parameters supplied by the user.
# One approach would be to use a templating system like Jinja,
# but this may be hard to maintain: The template itself is not Python,
# so syntax problems will not be obvious until it is filled in.
# At the other extreme, constructing code via an AST is very low-level.
#
# DP Wizard Templates is an alternative. The templates are themselves Python code,
# and the slots to fill are all-caps. This convention means that the template
# itself can be parsed as Python code, so syntax highlighting and linting still works.
#
#
# ## Examples: `dp_wizard_templates.code_template`
#
# There are two modules in this library. We'll look at `code_template` first.

# +

from dp_wizard_templates.code_template import Template


def conditional_print_template(CONDITION, MESSAGE):
    if CONDITION:
        print(MESSAGE)


conditional_print = (
    Template(conditional_print_template)
    .fill_expressions(CONDITION="temp_c < 0")
    .fill_values(MESSAGE="It is freezing!")
    .finish()
)

assert conditional_print == "if temp_c < 0:\n    print('It is freezing!')"

# -

# Note the different methods used:
# - `fill_expressions()` fills the slot with verbatim text.
#   It can be used for an expression like this, or for variable names.
# - `fill_values()` fills the slot with the repr of the provided value.
#   This might be a string, or it might be a list or dict or other
#   data structure, as long as it has a usable repr.
# - `finish()` converts the template to a string, and will error
#   if not all slots have been filled.
#
# Templates can also be in standalone files. If a `root` parameter is provided,
# the system will prepend `_` and append `.py` and look for a corresponding file.
# (The convention of prepending `_` reminds us that although these files
# can be parsed, they should not be imported or executed as-is.)

# +

from pathlib import Path

root = Path(__file__).parent.parent

block_demo = (
    Template("block_demo", root=root / "examples")
    .fill_expressions(FUNCTION_NAME="freeze_warning", PARAMS="temp_c")
    .fill_blocks(INNER_BLOCK=conditional_print)
    .finish()
)

assert (
    block_demo
    == '''def freeze_warning(temp_c):
    """
    This demonstrates how larger blocks of code can be built compositionally.
    """
    if temp_c < 0:
        print('It is freezing!')
'''
)

# -

# Finally, plain strings can also be used for templates.

# +

assignment = (
    Template("VAR = NAME * 2")
    .fill_expressions(VAR="band")
    .fill_values(NAME="Duran")
    .finish()
)

assert assignment == "band = 'Duran' * 2"

# -

# ## Examples: `dp_wizard_templates.converters`
#
# DP Wizard Templates also includes utilities to convert Python code
# to notebooks, and to convert notebooks to HTML. It is a thin wrapper
# which provides default settings for `nbconvert` and `jupytext`.
#
# The Python code is converted to a notebook using the
# [jupytext light format](https://jupytext.readthedocs.io/en/latest/formats-scripts.html#the-light-format):
# Contiguous comments are coverted to markdown cells,
# and contiguous lines of code are converted to code cells.
#
# One additional feature is that a section with a `# Coda` header
# will be stripped from notebook output. This allows a notebook
# to produce other artifacts without adding clutter.

# +

from dp_wizard_templates.converters import convert_py_to_nb, convert_nb_to_html


def notebook_template(TITLE, BLOCK, FUNCTION_NAME):
    # # TITLE
    #
    # Comments will be rendered as *Markdown*.
    # The `+` and `-` below ensure that only one code cell is produced,
    # even though the lines are not contiguous

    # +
    BLOCK

    FUNCTION_NAME(-10)
    # -

    # # Coda
    #
    # Extra computations that will not be rendered.

    2 + 2


title = "Hello World!"
notebook_py = (
    Template(notebook_template)
    .fill_blocks(BLOCK=block_demo)
    .fill_expressions(FUNCTION_NAME="freeze_warning", TITLE=title)
    .finish()
)

notebook_ipynb = convert_py_to_nb(notebook_py, title=title, execute=True)
(root / "examples" / "hello-world.ipynb").write_text(notebook_ipynb)

notebook_html = convert_nb_to_html(notebook_ipynb)
(root / "examples" / "hello-world.html").write_text(notebook_html)

# -

# The [output](examples/hello-world.html) is short,
# but it is an end-to-end demonstration of DP Wizard Templates,
# and as noted at the top, this documentation itself is rendered with
# `convert_py_to_nb` and `convert_nb_to_html`.
#
# ## Last thoughts
#
# Because the templates are valid Python, linters and other tools
# will by default include them in their coverage, and this may not be
# what you want. The exact configuration tweaks needed will depend
# on your tools, but here are some recommendations:
#
# - You might keep template files under `templates/` subdirectories,
#   and configure pytest (`--ignore-glob '**/templates/`) and pyright
#   (`ignore = ["**/templates/"]`) to ignore them.
# - For template functions, you might have a consistent naming
#   convention, and configure coverage (`exclude_also = def template_`)
#   to exclude them as well.

# # Coda

# +

readme_test_py = Path(__file__).read_text()

html_path = root / "index.html"
before_hash = hash(html_path.read_text())

html = convert_nb_to_html(
    convert_py_to_nb(readme_test_py, "DP Wizard Templates"), numbered=False
)
(html_path).write_text(html)

after_hash = hash(html_path.read_text())
assert (
    before_hash == after_hash
), "index.html has changed: If that is intended, the next test run should pass."
